// Copyright (C) 2008-2015 Tom Surace.

/**
 * @constant
 * @type Number
 */
var GuestUserId = 1;

/**
 * @constant
 * @type Number
 */
var AdminUserId = -1;

/**
 * Global (augh) event bus
 */
var eventBus = _.extend({}, Backbone.Events);

/**
 * @function
 * Helper that calls the ajax entry points for functions that don't
 * return data (ie, func/suckMore, etc), with standard error handling.
 * 
 * @param {!string} url location to GET
 * @param {Object} data Parameters for ajax call (data member of $.ajax).
 */
var getJson = function(url, data) 
{
    $.ajax({
        url: url,
        data: data,
        dataType: 'json',
        type: 'GET'
    });
};


var Mood = Backbone.Model.extend({
    defaults: {
        // id: null, // create new
        name: ""
    },
    
    urlRoot: '/func/mood'
});

/**
 * All available moods available.
 */
var MoodList = Backbone.Collection.extend({
    model: Mood,
    url: '/func/mood'
    
});

var User = Backbone.Model.extend({
    defaults: {
        // id: null, // create new
        name: "",
        password: "",
        role: "NOBODY"   // CAT MOUSE NOBODY
    },
    
    urlRoot: '/func/admin/user'
});

/**
 * All available users
 */
var UserList = Backbone.Collection.extend({
    model: User,
    url: '/func/admin/user'
    
});

/**
 * One line in the playback history (and pre-history) record. This is a 
 * read-only collection with no id.
 */
var HistoryInfo = Backbone.Model.extend({
    defaults: {
        key: 0,
        title: "",
        artist: "",
        album: "",
        suck: 50, // percent
        mood: 50, // percent
        status: 'UNKNOWN', // 'OK' 'MISSING'
        playCount: 0
    },

    urlRoot: '/func/history'
});

/**
 * @class
 */
var HistoryInfoList = Backbone.Collection.extend({
    model: HistoryInfo,
    url: '/func/history'
});

var CompressionSettings = Backbone.Model.extend({
    defaults: {
        name: "Default",
        ratio: 6,
        attack: 10,
        decay: 200,
        threshold: -30.0,
        gain: 10.0
    },
    
    urlRoot: '/func/compression'
});

var CompressionSettingsList = Backbone.Collection.extend({
    model: CompressionSettings,
    url: '/func/compression'
});

var SystemSettings = Backbone.Model.extend({
    defaults: {
        listenPort: 7076,
        mp3Root: ''
    },

    initialize: function() {
        this.set('compression', new CompressionSettings() );
    },
    
    parse: function(response){
        // Replace the "compression" field with a Compression model object.
        
        var data = response['compression'];
        response['compression'] = 
                new CompressionSettings( data, { parse: true });
        
        return response;
    },
    
    validate: function(attrs) {
        if (! attrs.mp3Root) {
            debugger;
            console.log( "mp3Root is not set: what happened?" );
            return "mp3Root is required in SystemSettings!";
        }
    },
    
    urlRoot: '/func/admin/settings'
});

/**
 * Represents and maps to the current status and the current user preferences.
 */
var Status = Backbone.Model.extend({
    defaults: {
        changeCount: -1,
        userId: GuestUserId,
        userName: '',
        role: "NOBODY",
        isCatIn: false,
        amIHidden: false,
        moodId: -1,
        moodName: '',
        nowPlaying: {
           key: -1
        },
        suckPercent: 0,
        moodPercent: 100,
        avgPlaycount: 0
    },

    /**
     * Don't include changeCount here so that model.fetch() forces a refetch
     * of current status regardless (post-login etc).
     */
    url: 'func/status',

    // Override fetch to ignore 304 "no change" (argh) which by definition
    // must contain no data, which freaks out the json response parser.
    fetch: function(options) {
        var thisPtr = this;

        $.ajax({
            type: 'GET',
            url: this.url,
            dataType: 'json',
            success: function(data, textStatus, jqxhr) { 
                if (jqxhr.status !== 304)
                    thisPtr.set( data );
            }
        });
        
        // Backbone.Model.prototype.fetch.apply( thisPtr, options );
    },
    
    /**
     * Sets the visiblity state on the server
     */
    setVisibility: function( amIVisible )
    {
        // This isn't a get, it's  a put!
        $.ajax({
            type: 'PUT',
            url: 'func/visible',
            dataType: 'json',
            data: {
                visible: amIVisible
            }
        });
    },

    /**
     */    
    suckLess: function()
    {
        getJson( 'func/status/suckLess', {
            trackId: this.get('nowPlaying').key
        });
    },
    
    /**
     * Tell the backend that the current track sucks more. If the current 
     * track (we think) is not what is current on the backend, there will be 
     * no effect.
     */
    suckMore: function()
    {
        getJson( 'func/status/suckMore', {
            trackId: this.get('nowPlaying').key
        });
    },

    megaSuck: function()
    {
        getJson( 'func/status/megaSuck', {
            trackId: this.get('nowPlaying').key
        });
    },

    moodYes: function()
    {
        var moodId = this.get('moodId');
        if (undefined === moodId || -1 === moodId) {
            alert("Current mood is invalid?"); // todo: global error event log
            return;
        }

        getJson( 'func/status/moodYes', {
                trackId: this.get('nowPlaying').key,
                moodId: moodId
        }); 
    },

    moodNo: function()
    {
        var trackId = this.get('nowPlaying').key;

        var moodId = this.get('moodId');
        if (undefined === moodId || -1 === moodId) {
            alert("Current mood is invalid?");
            return;
        }

        if (undefined === trackId) {
            alert("Track is invalid?");
            return;
        }

        getJson( 'func/status/moodNo', {
            trackId: trackId, 
            moodId: moodId
        });
    },
    
    /**
     * 
     * @param {type} moodId ID of the mood we want to use as the current one
     */
    moodSelect: function(moodId)
    {
        getJson( 'func/status/setMood', {
            moodId: moodId
        });
    },
    
    iAmACat: function() {
        return this.get('role') === 'CAT';
    }

});

var LoginView = Backbone.View.extend({

    initialize: function(){
        _.bindAll( this, "hide", "startLogin", "onIdClick" );

        this.$('#passwordBox').keyup(function(e){
            if (e.keyCode === 13)
                $(this).trigger('enter');
        });
    },
    
    events: {
        'enter': 'onIdClick',
        'click :visible #idBtn': 'onIdClick',
        'click :visible #cancel': 'hide'
    },
    
    hide: function()
    {
        this.$el.hide(125);
    },
    
    startLogin: function()
    {
        this.$el.show(125);
        
        this.$('#loginErrorBox').hide(125);
        this.$('#idBox').val('');

        // trigger "enter" events in the password box manually because
        // otherwise it won't work.
        
        this.$('#idBox').focus();
    },
    
    /**
     * Attempt to log in with the current credentials
     */
    onIdClick: function()
    {
        var id = this.$('#idBox').val();
        var pass = this.$('#passwordBox').val();
        $('#passwordBox').val(''); // Success or failure, you gotta retype it

        // TODO: some sort of minimal password hashing?

        var thisPtr = this;
        $.ajax({
            url: 'func/login',
            data: { 
                id: id, 
                password: pass
            },
            dataType: 'json',
            type: 'GET',
            success: function(data, textStatus, jqxhr) { 
                thisPtr.hide(125);
                thisPtr.trigger( 'loginSuccess' );
            },
            error: function(jqxhr, textStatus, errorThrown) { 
                thisPtr.loginFail( jqxhr, 'login', errorThrown ); 
            }
        });
    },
    
    loginFail: function( jqxhr, textStatus, errorThrown ){
        this.$('#loginErrorBox').show(125);
        this.$('#loginErrorBox').html( jqxhr.responseText );
        this.$('#idBox').focus();
    }
});

/**
 * Inline mood select/editor for the ControlsBox
 * Generates "mood:select" event when mood is selected.
 * model should be a MoodList collection.
 */
var MoodListView = Backbone.View.extend({

    initialize: function(){
        // _.bindAll( this, "hide", "startLogin", "onIdClick" );

        // We really only care about added/removed/modified Mood entries
        // if they have been saved to the server.
        this.model.on( 'sync', this.render, this );
        this.model.on( 'destroy', this.render, this );
        
        /**
         * @type {!boolean} If true, we show cat-only edit controls.
         */
        this.enableEdit = false;

        /**
         * @type {!boolean} If true, enable mood selection.
         */
        this.enableSelect = false;

        // Bind enter key from new and edit edit boxes
        this.$('#newName').keyup(function(e){
            if (e.keyCode === 13)
                $(this).trigger('enter:newName');
        });
        
        this.$('#editName').keyup(function(e){
            if (e.keyCode === 13)
                $(this).trigger('enter:editName');
        });
    },
    
    events: {
        // Mood form controls (should these even be here?)
        'click #select': 'onSelect',
        'dblclick #moodList': 'onSelect',
        'click #new': 'onNew',
        'click #newSubmit': 'onNewSubmit',
        'enter:newName': 'onNewSubmit',
        'click #edit': 'onEdit',
        'click #editSubmit': 'onEditSubmit',
        'enter:editName': 'onEditSubmit',
        'click #delete': 'onDelete',
        'click #cancel': 'onCancel',
        'change #moodListEdit': 'onMoodListEditChange'
    },
    
    hide: function()
    {
        this.$el.hide(125);
    },

    show: function()
    {
        this.$el.show(125);
        
        this.hideAllPages();
        this.$('#listPage').show();
        
        var selectBox = this.$('#moodList');
        selectBox.html('<option value="1">(Loading)</option>');
        selectBox.attr('disabled',true);
        
        // Refetch whenever we show it because, well, what if we logged in?
        this.model.fetch(); // update options
    },
    
    toggle: function()
    {
        if (this.$el.is(':visible'))
            this.hide();
        else
            this.show();
    },
    
    hideAllPages: function()
    {
        this.$('#listPage').hide();
        this.$('#newPage').hide();
        this.$('#editPage').hide();
    },
    
    /**
     */
    render: function() 
    {
        // Enable these once the list is loaded
        this.$('#select').attr('disabled',true);
        this.$('#delete').attr('disabled',true);
        
        this.$('#new').attr('disabled',!this.enableEdit);
        this.$('#delete').attr('disabled',!this.enableEdit);
        
        var options = '';
        for (var i = 0; i < this.model.length; i++)
        {
           var mood = this.model.at(i);
           options += '<option value="' + mood.id + '"';
           options += ">" + mood.get('name') + '</option>';
        }
    
        var ctrl = this.$('#moodList');
        ctrl.html( options );
        ctrl.attr('disabled', false);

        this.$('#delete').attr('disabled', !this.enableEdit);
        this.$('#select').attr('disabled', !this.enableSelect);
    },

    onSelect: function()
    {
        var moodId = this.$('#moodList').val();
        if (!moodId) 
        {
            console.log( "Mood ID is not valid: " + moodId );
            return;
        }

        this.hide();
        this.trigger( 'moodSelect', moodId );
    },
    
    onEdit: function()
    {
        var selectedId = this.$('#moodList').val();
        if (!selectedId) 
        {
            // LOG: nothing is selected?
            return;
        }
        var selectedText = $("#moodList option:selected" ).text();
        
        this.hideAllPages();
        this.$('#editPage').show();
        this.$('#editId').val(selectedId);
        this.$('#editName').val(selectedText); // empty form
        this.$('#editName').focus();
    },
    
    onEditSubmit: function()
    {
        var thisPtr = this;
        
        var id = this.$('#editId').val();
        var newName = this.$('#editName').val();
        if (!newName)
        {
            return;
        }
        
        var mood = this.model.get( id );
        mood.set({ name: newName });
        mood.save();
        
        this.hideAllPages();
        this.$('#listPage').show();
    },
    
    onNew: function()
    {
        this.hideAllPages();
        this.$('#newPage').show();
        this.$('#newName').val(''); // empty form
        this.$('#newName').focus();
    },
    
    onNewSubmit: function()
    {
        var thisPtr = this;
        
        var newName = this.$('#newName').val();
        if (!newName)
        {
            return;
        }
        var mood = new Mood({ name: newName });
        
        // Set just-created mood as current when we know its id
        // this.model.create( mood, {async: false} ); 
        this.model.create( mood, {
            success: function() {
                thisPtr.trigger( 'moodSelect', mood.get('id') );
            }
        });
        
        this.hide();
        
        // Select the new mood?
        // this.trigger( 'moodSelect', mood.get('id') );
    },
    
    onDelete: function()
    {
        var moodId = this.$('#moodList').val();
        if (!moodId) 
        {
            alert("Mood ID is not valid: " + moodId );
            return;
        }
        
        this.model.get( moodId ).destroy();
    },
    
    onCancel: function()
    {
        this.hide();
    },
    
    onMoodListChange: function()
    {
        // We are creating a new mood here. Deselect the existing one and
        // enable the Create button.
    }
});

var NewUserView = Backbone.View.extend({
    
    initialize: function()
    {
    },
    
    events: {
        'click .save':    'onSave',
        'click .cancel':  'onCancel'
    },
    
    show: function() 
    {
        this.model = new User();
        
        this.$el.show();
        this.$('.nameEdit').val(this.model.get('name'));
        this.$('.passwordEdit').val('');
      
        var role = this.model.get('role');
        this.$('input:radio[name="role"]').
                filter('[value="' + role + '"]').prop('checked',true);
        
        this.$('.nameEdit').focus();
    },
    
    hide: function() 
    {
        this.$el.hide();
    },
    
    onSave: function() 
    {
        // Get values from form, save
        var name = this.$('.nameEdit').val();
        var password = this.$('.passwordEdit').val();
        var role = $('input:radio[name=role]:checked').val();
        
        if (!name)
        {
            alert( "Name is required" );
            return;
        }
        if (!role)
        {
            alert( "Could not retrieve role from form" ); // internal error
            return;
        }
        
        this.model.set( { name: name,
                          password: password,
                          role: role } );
        
        this.model.save();
        
        Backbone.history.navigate( "#user", {trigger: true} );
    },
    
    onCancel: function()
    {
        Backbone.history.navigate( "#user", {trigger: true} );
    }
});

var EditUserView = Backbone.View.extend({
    
    initialize: function()
    {
    },
    
    events: {
        'click .save':    'onSave',
        'click .cancel':  'onCancel'
    },

    /**
     * 
     * @param {User} user
     */
    show: function( user )
    {
        this.stopListening(); // disconnect any old model event handlers
        
        this.model = user;

        this.$('.nameEdit').val('[loading]');
        this.$('.passwordEdit').val('');
      
        // Show, but disabled, while we wait for fetch
        this.$el.show(); 
        this.$('input').attr( 'disabled', true );
        
        this.listenTo( this.model, "change", this.render );
        this.model.fetch(); // get latest values
    },

    render: function()
    {
        
        this.$('.nameEdit').val(this.model.get('name'));
        this.$('.passwordEdit').val('');
      
        var role = this.model.get('role');
        this.$('input:radio[name="role"]').
                filter('[value="' + role + '"]').prop('checked',true);
        
        // Now that values are loaded: enable and focus
        this.$('input').attr( 'disabled', false );
        this.$('.nameEdit').focus();
    },
    
    hide: function() 
    {
        this.stopListening();
        this.$el.hide();
    },
    
    onSave: function() 
    {
        // Get values from form, save
        var name = this.$('.nameEdit').val();
        var password = this.$('.passwordEdit').val();
        var role = $('input:radio[name=role]:checked').val();
        
        if (!name)
        {
            alert( "Name is required" );
            return;
        }
        if (!role)
        {
            alert( "Could not retrieve role from form" ); // internal error
            return;
        }
        
        this.model.set( { name: name,
                          password: password,
                          role: role } );
        
        this.model.save();
        
        Backbone.history.navigate( "#user", {trigger: true} );
    },
    
    onCancel: function()
    {
        Backbone.history.navigate( "#user", {trigger: true} );
    }
});

var UserListView = Backbone.View.extend({

    initialize: function()
    {
        this.model.on( 'sync', this.render, this );
        this.model.on( 'destroy', this.render, this );
    },

    events: {
        // Mood form controls (should these even be here?)
        'click .new ': 'onListNew',
        'click .edit ': 'onListEdit',
        'dblclick .userList': 'onListEdit',
        'click .delete ': 'onListDelete'
    },
    
    render: function()
    {
        var options = '';
        for (var i = 0; i < this.model.length; i++)
        {
            var userInfo = this.model.at(i);
 
            var userName = userInfo.get('name');
            var id = userInfo.get('id');

            var cat = '';
            if (userInfo.get('role') === 'CAT')
                cat = ' [CAT]';

            options += '<option value="' + id + '">' 
                     + userName + cat + '</option>';
        }
    
        var ctrl = this.$('.userList');
        ctrl.html( options );
    },
    
    onListNew: function()
    {
        Backbone.history.navigate( "#user/new", {trigger: true} );
    },
    
    onListEdit: function()
    {
        var id = this.getSelectedUser();
        if (id)
            Backbone.history.navigate( "#user/edit/" + id, {trigger: true} );
    },
    
    onListDelete: function()
    {
        // TODO: confirm?
        var id = this.getSelectedUser();
        if (id)
        {
            var user = this.model.get(id);
            if (user)
            {
                user.destroy();
            }
        }
    },
    
    /**
     * 
     * @returns {number} id of the current selection in the list, or undefined
     *   if nothing is selected
     */
    getSelectedUser: function()
    {
        return this.$('.userList').find(':selected').val();
    },
    
    show: function()
    {
        this.$el.show();
        this.model.fetch();
    },
       
    hide: function()
    {
        this.$el.hide();
    }
});

var CompressionPresetsView = Backbone.View.extend({

    initialize: function(options)
    {
        // this.model is a CompressionSettingsList
        
        // The global SysetmSettings object
        this.systemSettings = options.systemSettings;
        
        this.model.on( 'sync', this.render, this );
        this.model.on( 'destroy', this.render, this );
    },
    
    events: {
        'click .select ': 'onListSelect',
        'dblclick .presetList': 'onListSelect',
        'click .create ': 'onCreate',
        'click .cancel ': 'onCancel',
        'click .delete ': 'onListDelete'
    },
    
    render: function()
    {
        var options = '';
        for (var i = 0; i < this.model.length; i++)
        {
            var settings = this.model.at(i);
 
            var name = settings.get('name');
            var id = settings.get('id');

            options += '<option value="' + id + '">' + name + '</option>';
        }
    
        var ctrl = this.$('.presetList');
        ctrl.html( options );
    },
    
    show: function()
    {
        this.$el.show();
        this.model.fetch();
        this.$('.presetList').focus();
    },
       
    getSelected: function()
    {
        return this.$('.presetList').find(':selected').val();
    },
    
    hide: function()
    {
        this.$el.hide();
    },
    
    onCancel: function()
    {
        Backbone.history.navigate( "#audioSettings", {trigger: true} );
    },
    
    /** 
     * Create a new preset with default values and return to the
     * main audio edit screen.
     */
    onCreate: function()
    {
        var newSettings = new CompressionSettings( { name: 'New Settings' } );
        this.systemSettings.set( 'compression', newSettings );
        this.systemSettings.save();
        Backbone.history.navigate( "#audioSettings", {trigger: true} );
    },
    
    onListSelect: function()
    {
        var id = this.getSelected();
        if (id)
        {
            var compressionSettings = this.model.get( id );
            this.systemSettings.set( 'compression', compressionSettings );
            this.systemSettings.save();
            Backbone.history.navigate( "#audioSettings", {trigger: true} );
        }
    },
    
    onListDelete: function()
    {
        var id = this.getSelected();
        if (! id) {
            return;
        }

        var preset = this.model.get(id);
        if (!preset) {
            return;
        }

        preset.destroy();
        
        // If the preset we destroyed is also the currently active
        // preset, select another at random. 
        // If none are left, create one and use it.

        var currentModel = this.systemSettings.get( 'compression' );
        if (currentModel.get('id') == id)
        {
            if (this.model.length <= 0) 
            {
                // Delegate creation (includes navigation, so return)
                this.onCreate();
                return;
            }
            else
            {
                var use = this.model.at( 0 ); // first in list
                this.systemSettings.set( 'compression', use );
                this.systemSettings.save();
            }
        }
        
        Backbone.history.navigate( "#audioSettings", {trigger: true} );
    }    
   
});

var AudioSettingsView = Backbone.View.extend({

    initialize: function(options){
        this.listenTo( this.model, "sync", this.render );
        
        // Save settings on enter, because that's convenient when editing
        this.$('.nameEdit, .radioEdit, .attackEdit, .decayEdit, .thresholdEdit, .gainEdit').keyup(function(e){
            if (e.keyCode === 13)
                $(this).trigger('enter');
        });
    },
    
    events: {
        'click .save':    'onSave',
        'enter':          'onSave',
        'click .presets': 'onShowPresets'
    },
    
    show: function() 
    {
        this.$el.show();
        this.$('input').attr( 'disabled', true );
        this.model.fetch();
    },
    
    hide: function() 
    {
        this.$el.hide();
    },
    
    render: function() {
        var compModel = this.model.get('compression');
        this.$('.nameEdit').val( compModel.get('name') );
        this.$('.ratioEdit').val( compModel.get('ratio') );
        this.$('.attackEdit').val( compModel.get('attack') );
        this.$('.decayEdit').val( compModel.get('decay') );
        this.$('.thresholdEdit').val( compModel.get('threshold') );
        this.$('.gainEdit').val( compModel.get('gain') );
        this.$('input').attr( 'disabled', false );
        
        this.$('.nameEdit').focus();
    },
    
    onShowPresets: function() {
        Backbone.history.navigate( "#compression/list", {trigger: true} );
    },
    
    onSave: function() {
        var name = this.$('.nameEdit').val();
        var ratio = this.$('.ratioEdit').val();
        var attack = this.$('.attackEdit').val();
        var decay = this.$('.decayEdit').val();
        var threshold = this.$('.thresholdEdit').val();
        var gain = this.$('.gainEdit').val();
        
        // VALIDATE
        if (isNaN(parseFloat(ratio)) || !isFinite(ratio))
            ratio = 8;
        else if (ratio < 1)
            ratio = 1;
        else if (ratio > 9999)
            ratio = 9999;
        
        
        if (attack < 0)
            attack = 0;
        
        if (decay < 0.010)
            decay = 0.010;

        if (threshold < -120)
            threshold = -120;
        
        if (threshold > 6)
            threshold = 6;
        
        if (gain < -100)
            gain = -100;
        
        if (gain > 100)
            gain = 100;
        
        this.$('.ratioEdit').val( ratio );
        this.$('.attackEdit').val( attack );
        this.$('.decayEdit').val( decay );
        this.$('.thresholdEdit').val( threshold );
        this.$('.gainEdit').val( gain );
        
        var compModel = this.model.get( 'compression' );
        compModel.set( { 
            'name': name,
            'ratio': ratio,
            'attack': attack,
            'decay': decay,
            'threshold': threshold,
            'gain': gain
        } );
        
        this.model.save();
    }
});

var SystemSettingsView = Backbone.View.extend({

    initialize: function(options){
        this.listenTo( this.model, "sync", this.render );
    },
    
    events: {
        'click .save':    'onSave'
    },
    
    show: function() 
    {
        this.$el.show();
        this.$('input').attr( 'disabled', true );
        this.model.fetch();
    },
    
    hide: function() 
    {
        this.$el.hide();
    },
    
    render: function() {
        this.$('.mp3RootEdit').val( this.model.get('mp3Root') );
        
        this.$('input').attr( 'disabled', false );
    },
    
    onSave: function() {
        var root = this.$('.mp3RootEdit').val();
        
        // Validate!
        if (root === '')
        {
            return; // We're not saving that.
        }
        
        this.model.set({ 'mp3Root': root });
        this.model.save();
        Backbone.history.navigate( "", {trigger: true} );
    }
});

/**
 * Construct with model=Status model, and moodlist=MoodList collection
 */
var ControlsBoxView = Backbone.View.extend({

    initialize: function(options){
        _.bindAll( this, "render", "hide", "show", "onSuck", "onRule", "onMegaSuck", "onHideMeBox" );
        
        this.model.on( 'change', this.render, this );
        
        // The child moodlist view 
        this.moodListView = new MoodListView({ 
            el: this.$('#moodForm'),
            model: options.moodList });
        
        this.moodListView.on( 'moodSelect', this.moodSelectHandler, this );
    },
    
    events: {
        // Only register clicks on visible controls!
        'click :visible #login': 'onLogin',
        'click :visible #hideMeBox': 'onHideMeBox',
        'click :visible #suckBtn': 'onSuck',
        'click :visible #ruleBtn': 'onRule',
        'click :visible #megaSuckBtn': 'onMegaSuck',
        'click :visible #yesBtn': 'onYes',
        'click :visible #noBtn': 'onNo',
        'click :visible #moodBtn': 'onMoodClick'
    },
   
    hide: function()
    {
        this.$el.hide(125);
    },
    
    show: function()
    {
        this.$el.show(125);
    }, 
    
    setAnonymousControlsDisabled: function( isDisabled )
    {
        this.moodBtn.attr( 'disabled', isDisabled );
        this.ruleBtn.attr( 'disabled', isDisabled );
        this.suckBtn.attr( 'disabled', isDisabled );
        this.megaSuckBtn.attr( 'disabled', isDisabled );
        this.yesBtn.attr( 'disabled', isDisabled );
        this.noBtn.attr( 'disabled', isDisabled );
    },
    
    onHideMeBox: function()
    {
        var hideMe = this.$('#hideMeBox').is(':checked');
        this.model.setVisibility( !hideMe );
    },
    
    onLogin: function()
    {
        // Publish an event on the controls box view
        this.trigger( 'click:login' );
    },
   
    onRule: function()
    {
        this.setProgress( "Decreasing Suck for ("
                          + this.model.get('nowPlaying').key
                          + "): " 
                          + this.model.get('nowPlaying').title );

        this.model.suckLess();
    },

    /**
     * Callback for the Main Suck Button
     */
    onSuck: function()
    {
        this.setProgress( "Increasing Suck for What You Hear (" 
                          + this.model.get('nowPlaying').key
                          + "): " 
                          + this.model.get('nowPlaying').title);
                  
        this.model.suckMore();
    },
    
    onMegaSuck: function()
    {
        this.setProgress( "Mega Suck!" );
        this.model.megaSuck();
    },

    onYes: function()
    {
        var nowPlaying = this.model.get('nowPlaying');
        this.setProgress( "Yes! (" 
                          + nowPlaying.key
                          + "): " 
                          + nowPlaying.title );

        this.model.moodYes();
    },

    onNo: function()
    {
        this.setProgress( "No! (" + this.model.get('nowPlaying').key + "): " 
                          + this.model.get('nowPlaying').title );

        this.model.moodNo();
    },
    
    /**
     * Click handler for the "start selecting moods" button.
     * On click, unhide the mood selection region and fetch() current moods
     */
    onMoodClick: function()
    {
        var thisPtr = this;
        
        var moodForm = this.$('#moodForm');
        
        if (moodForm.is(':visible'))
        {
            moodForm.hide(125);
            return;
        }
        
        this.moodListView.toggle();
    },
    
    moodSelectHandler: function(moodId)
    {
        // Select this mood in the current status.
        this.model.moodSelect(moodId);
    },

    setProgress: function(msg) {
        // TODO: replace with a view that slowly replaces status over time?
        this.$('#progress').html( msg );
    },

    render: function() {
        // Cached jquery objects to update
        if (!this.elementsAreCached)
        {
            // Should private elements use _ prefix? Should be marked private
            // in javaDoc? Yes.
            this.elementsAreCached = true;
            this.loginBtn = this.$('#login');
            this.moodBtn = this.$('#moodBtn');
            this.role = this.$('#role');
            this.catStat = this.$('#catStat');
            this.trackId = this.$('#trackId');
            this.playCount = this.$("#playCount");
            this.title = this.$("#title");
            this.artist = this.$("#artist");
            this.album = this.$("#album");
            this.filename = this.$("#filename");
            this.suckLevel = this.$("#suckLevel");
            this.moodLevel = this.$("#moodLevel");
            this.avgPlayCount = this.$('#avgPlayCount');
            this.hideMeArea = this.$('#hideMeArea');
            this.hideMeBox = this.$('#hideMeBox');
            this.suckBtn = this.$('#suckBtn');
            this.ruleBtn = this.$('#ruleBtn');
            this.megaSuckBtn = this.$('#megaSuckBtn');
            this.yesBtn = this.$('#yesBtn');
            this.noBtn = this.$('#noBtn');
        }
        
        var status = this.model; // local for simplicity.

        var userId = status.get('userId');
        if (!userId || GuestUserId === userId)
        {
           this.loginBtn.val("Log In");
           this.moodBtn.val( status.get('moodName') );
           this.setAnonymousControlsDisabled( true );
        }
        else
        {
           this.loginBtn.val( status.get('userName') );
           this.moodBtn.val( status.get('moodName') );
           this.setAnonymousControlsDisabled( false );
        }

        this.role.html( status.get('role') );
        var isCatIn = status.get('isCatIn');
        this.catStat.html( isCatIn ? "in" : "away" );
        
        this.trackId.html( status.get('nowPlaying').key );
        this.playCount.html( status.get('nowPlaying').playCount );
        this.title.html( status.get('nowPlaying').title );
        this.artist.html( status.get('nowPlaying').artist );
        this.album.html( status.get('nowPlaying').album );
        this.filename.html( status.get('nowPlaying').filePath );
        this.suckLevel.html( status.get('suckPercent') );
        this.moodLevel.html( status.get('moodPercent') );
        this.avgPlayCount.html( status.get('avgPlayCount') );

        if (status.iAmACat())
        {
           this.hideMeArea.css('display', 'inline');
           this.hideMeBox.attr('checked', status.get('amIHidden'));
        }
        else
        {
           // Set the CSS instead of using show() and hide() because
           // this is supposed to be an inline element.
           this.hideMeArea.css('display', 'none');
        }
        
        this.moodListView.enableEdit = status.iAmACat();
        this.moodListView.enableSelect = status.iAmACat() || !status.get('isCatIn');

        return this;
    }
});

var HistoryView = Backbone.View.extend(
/** @lends HistoryView.prototype */
{
    events: {
        'click a.suckLess ': 'onSuckLess',
        'click a.suckMore ': 'onSuckMore',
        'click a.moodLess ': 'onMoodLess',
        'click a.moodMore ': 'onMoodMore'
    },
    
    initialize: function() {
        this.model.on( 'sync', this.render, this );
    },
    
    hide: function() {
        this.$el.hide();
    },
    
    show: function() {
        this.$el.show();
        this.model.fetch();
    },
    
    onSuckLess: function(e) {
        e.preventDefault();
        var id = $(e.currentTarget).data("key");
        getJson( 'func/history/suckLess', {
            trackId: id
        });
    },
    
    onSuckMore: function(e) {
        e.preventDefault();
        var id = $(e.currentTarget).data("key");
        getJson( 'func/history/suckMore', {
            trackId: id
        });
    },
    
    onMoodLess: function(e) {
        e.preventDefault();
        var id = $(e.currentTarget).data("key");
        getJson( 'func/history/moodNo', {
            trackId: id
        });
    },
    
    onMoodMore: function(e) {
        e.preventDefault();
        var id = $(e.currentTarget).data("key");
        getJson( 'func/history/moodYes', {
            trackId: id
        });
    },
    
    /**
     * @augments external:Backbone.View
     */
    render: function() {
        if( !this.$el.is(":visible") ){  // Not visible?
            return this; // ** quick exit **
        }
        
        var data = { "data": this.model.toJSON() };
        var output = _.template( $("#historyList_tmpl").html(), data );
        this.$('#historyBox').html( output );
        
        return this;
    }
});

// 
// 
//
// Master view for the page. I guess
//
var MasterView = Backbone.View.extend(
/** @lends MasterView.prototype */
{
    /**
     * @augments external:Backbone.View
     * @constructs
     * @param {Object} options
     */
    initialize: function(options) {
        _.bindAll( this, "startLogin" );
        
        this.controlsBoxView = new ControlsBoxView({ 
            el: '#controlsBox',
            model: options.statusModel,
            moodList: options.moodList
        });

        this.historyView = new HistoryView({ 
            el: '#historyView',
            model: options.historyInfoList,
            statusModel: options.statusModel
        });
        
        this.systemSettingsModel = new SystemSettings();
        
        this.audioSettingsView = new AudioSettingsView({
            el: '#audioSettingsView .mainPage',
            model: this.systemSettingsModel
        });
        
        this.compressionPresetsView = new CompressionPresetsView({
            el: '#audioSettingsView .presetsPage',
            model: new CompressionSettingsList(),
            systemSettings: this.systemSettingsModel
        });
        
        this.systemSettingsView = new SystemSettingsView({
            el: '#systemSettingsView',
            model: this.systemSettingsModel
        });
        
        // Just the list part of the user editor
        this.userListView = new UserListView({
            el: '#usersView .listPage',
            model: options.userList
        });
        
        this.newUserView = new NewUserView({
            el: '#usersView .newPage'
        });
        
        this.editUserView = new EditUserView({
            el: '#usersView .editPage'
        });
        
        this.loginView = new LoginView({ 
            el: '#loginForm'
        });
        
        
        // May wish to create a publish subscribe setup (or use routers)
        // later, but for now just bind interesting events to other views
        // here.
        
        this.controlsBoxView.on('click:login', this.startLogin, this );
        
        this.loginView.hide(125);
    },
    
    startLogin: function()
    {
        this.loginView.startLogin();
    },
    
    hideDetailPages: function()
    {
        this.historyView.hide();
        
        this.userListView.hide();
        this.newUserView.hide();
        this.editUserView.hide();
        $('#usersView').hide();
        $('#infoView').hide();
        $('#audioSettingsView').hide();
        
        this.audioSettingsView.hide();
        this.systemSettingsView.hide();
        this.compressionPresetsView.hide();
    },
    
    showDefaultView: function()
    {
        // navigate to the home view == history
        this.showHistoryView();
    },
    
    showCompressionPresets: function()
    {
        this.hideDetailPages();
        $('#audioSettingsView').show();
        this.compressionPresetsView.show();
    },
    
    showHistoryView: function()
    {
        this.hideDetailPages();
        this.historyView.show();
    },
    
    showAudioSettings: function()
    {
        this.hideDetailPages();
        $('#audioSettingsView').show();
        this.audioSettingsView.show();
    },
    
    showSystemSettings: function()
    {
        this.hideDetailPages();
        this.systemSettingsView.show();
    },
    
    showInfo: function()
    {
        this.hideDetailPages();
        $('#infoView').show();
    },

    listUsers: function()
    {
        this.hideDetailPages();
        $('#usersView').show();
        this.userListView.show();
    },
    
    newUser: function()
    {
        this.hideDetailPages();
        $('#usersView').show();
        this.newUserView.show();
    },
    
    editUser: function( id )
    {
        this.hideDetailPages();
        $('#usersView').show();
        
        var user = new User({ id: id });
        user.fetch();
        if (!user)
        {
            console.log( "Can not edit user " + id + ". Not found!" );
            return;
        }
        
        this.editUserView.show( user );
    }
    
    /**
     * @augments external:Backbone.View
     */
    /*
    render: function() {
        this.controlsBoxView.render();
        this.historyView.render(); // IS VISIBLE?
        return this;
    }
    */

});

var JBRouter = Backbone.Router.extend({
    
    initialize: function(options)
    {
        this.status = new Status();
        this.moodList = new MoodList();
        this.historyInfoList = new HistoryInfoList();
        this.userList = new UserList();

        this.masterView = new MasterView({ 
            statusModel: this.status,
            moodList: this.moodList,
            historyInfoList: this.historyInfoList,
            userList: this.userList
        });

        // Set initial view values
        this.status.set({
            changeCount: -1,
            userId: GuestUserId,
            userName: '',
            role: "LOADING",
            isCatIn: false,
            amIHidden: false,
            moodId: -1,
            moodName: 'Unknown',
            nowPlaying: {
               key: -1
            },
            suckPercent: 0,
            moodPercent: 100,
            avgPlaycount: 0
        });
        
        /**
         * If true, it forces a refresh of the status, for example after
         * successful login when the old credentials (passed in to the call)
         * are no longer valid. Reset after every fetch.
         */
        this.ignoreNextStatus = false;
        this.masterView.loginView.on('loginSuccess', this.forceGuiRefresh, this );

        this.slowFetchStatus();
    },

    routes: {
      "history":              "history", // #history
      "info":                 "info",    // #info
      "audioSettings":        "audioSettings",
      "compression/list":     "compressionPresets",
      "systemSettings":       "systemSettings",
      "user":                 "users",   // #user
      "user/new":             "userNew",
      "user/edit/:id":        "userEdit",
      "":                     "home",
      "*path":                "defaultRoute" // Wildcard route
    },

    defaultRoute: function(path) {
        console.log( "Route not handled: " + path );
    },

    audioSettings: function() {
        this.masterView.showAudioSettings();
    },
    
    compressionPresets: function() {
        this.masterView.showCompressionPresets();
    },
    
    home: function(){
        this.masterView.showDefaultView();
    },

    history: function() {
        this.masterView.showHistoryView();
    },
    
    info: function() {
        this.masterView.showInfo();
    },

    systemSettings: function() {
        this.masterView.showSystemSettings();
    },
    
    users: function() {
        this.masterView.listUsers();
    },
    
    userNew: function() {
        this.masterView.newUser();
    },
    
    userEdit: function(id) {
        this.masterView.editUser(id);
    },

    /**
     * forces a full refresh of the controls box, and cancels any pending
     * calls. Useful after a login when any overlapping requests could crush
     * our login cookies.
     */
    forceGuiRefresh: function()
    {
        this.cancelSlowFetch();
        this.status.fetch();
        this.historyInfoList.fetch();
    },
    
    cancelSlowFetch: function()
    {
        if (this.currentStatusAjax)
        {
            try
            {
                this.currentStatusAjax.abort();
            }
            catch (ex) { /* ignore */ }
        }
    },
    
    slowFetchStatus: function()
    {
        var thisPtr = this;

        this.currentStatusAjax = 
            $.ajax({
                url: 'func/status',
                data: { 
                    changeCount: this.status.get('changeCount') 
                },
                dataType: 'json',
                type: 'GET',
                timeout: 300000, // 5m
                success: function(data, textStatus, jqxhr) { 
                    var delay = 2000;
                    if (jqxhr.status === 304) // timeout? 
                    {
                        // TODO: anything special at all?
                    }
                    else // GOT NEW STATUS! (also let's update the history)
                    {
                        try
                        {
                            if (thisPtr.ignoreNextStatus) // Fetch immediately
                            {
                                thisPtr.ignoreNextStatus = false;
                            }
                            else
                            {
                                // TODO: publish as a status update event,
                                // handle things like login-timeout etc in the
                                // main router object.
                                
                                thisPtr.status.set( data );
                                thisPtr.historyInfoList.fetch();
                            }
                        }
                        finally // no mistakes: keep the fetch running! 
                        {
                            // Tuning: I find that it usually takes about two 
                            // seconds after a track changes for the queue to
                            // be completely filled again. ~2s delay should 
                            // mean there's only one extra call to fully update
                            // the queue.
                            // delay = 2000;
                            
                            // Low latency: we don't care about multiple refresh
                            // calls because we don't
                            delay = 100;
                        }
                    }

                    window.setTimeout( 
                        function(){thisPtr.slowFetchStatus();}, delay );
                },
                error: function( jxhqr, textStatus, errorThrown) 
                {
                    var delay = 2000;
                    if (textStatus === "timeout" || textStatus === "abort")
                    {
                        // It's OK, we expect these
                        delay = 2000;
                    }
                    else
                    {
                        console.log( "Status fetch error: " + textStatus );
                        delay = 60000;
                    }
                    window.setTimeout( 
                        function(){thisPtr.slowFetchStatus();}, delay );
                }
            });
    }
});

/**
 * @constructor
 */  
function index()
{
    this.router = new JBRouter({});
    Backbone.history.start(); // Start the router
};

